// Author: George W. Taylor
// http://www.tropicalcoder.com/
// Date: 15 Oct 2012
// License: Creative Commons - Non-Commercial Use

#include "stdafx.h"
#include <math.h>

#include "ASIOSupport.h"
#include "CParams.h"
#include "Debug.h"


#define MAX_16BIT_AMPLITUDE 0x00007FFF
#define MAX_16BIT_DBS ((double)90.308733622833978944959868280967)
#define TWO_PI ((double)6.2831853071795865)
#define NUM_STEREO_CHANNELS_REQD 1


void ASIO_IO::KeyOn(void)
{
 if(bKeyedOn)return; // Already on!

 // Did we get a key on command too soon,
 // while it is still in either the attack or decay phase?
 // Then force it on!  (We don't expect this to happen normally)
 if(bRising || bFalling)
 {
	bRising = bFalling = bCheckSampleCount = false;
    bKeyedOn = true;
    dwCountSamples = 0;
	return;
 }

 // Trigger attack...
 dPhaseAngle = 0; // Start each tone at zero phase
 dwCountSamples = 0;
 dwCountRiseTimeSamples = 0;
 dwCountFallTimeSamples = 0;
 bKeyedOn = bRising = true;
 bCheckSampleCount = bFalling = false;
}

void ASIO_IO::KeyOff(DWORD dwDurationMicroseconds)
{
 if(!bKeyedOn)return; // Already off!

 // Did we get a key off command too soon,
 // while it is still in either the attack or decay phase?
 // Then force if off! (We don't expect this to happen normally)
 if(bRising || bFalling)
 {
    bKeyedOn = false;
    bRising = bCheckSampleCount = bFalling = false;
	return;
 }

 // We have the duration a tone was keyed on in microseconds.
 // What is that in samples?
 dwTotalDurationSamples = (DWORD)floor(((double)dwDurationMicroseconds / dSampleMicroseconds) + 0.5);

 // We only care about dwTotalDurationSamples when we go to stop the tone,
 // and find we are short some samples according to sample duration commanded.
 // In that case, we will keep generating samples until we reach duration commanded,
 // before wrapping up by adding any decay delay and continuing to the nearest full wave zero crossing.
 // However, since continuing to the nearest full wave zero crossing will add from 0 to almost 1 full wave,
 // we will only add samples to compensate for timing errors up to 1/2 wavelength short of dwTotalDurationSamples,
 // so that after continuing to the nearest full wave zero crossing, the final total will be +/- 1/2 wavelength,
 // rather than 0 to 1 wavelengths.
 // Then reduce our target sample count by 1/2 wavelength...
 if(dwTotalDurationSamples > dwHalfIntegralWaveLength) // avoid rollover error, however unlikely!
    dwTotalDurationSamples -= dwHalfIntegralWaveLength;

 dwCountFallTimeSamples = 0;
 bFalling = bRising = false;
 // Trigger the beginning of the end...
 bCheckSampleCount = true;
}


// Called when (bFalling && dwCountFallTimeSamples == 0)
// Add decay delay sample count if any
// Determine Num sample to next full wave zero crossing
// Store that value in dwTotalFallTimeSamples
void CalculateTotalFallTimeSamples(ASIO_IO *pObj)
{
 double dWavFrac;
 DWORD dwSamplesToZeroCrossing = 0;

 if(pObj->bUsingDecayDelay)
 {
    // Trigger decay, ending at a zero crossing...
    // Num fall time samples as per user request: (CParams::dFallTime * dwSampleRate)
    // Num sample to next full wave zero crossing: fmod((double)dwCountSamples, dWaveLength)
    // Total fall time samples = dwCalculatedFallTimeSamples + dwSamplesToZeroCrossing (Num sample to next full wave zero crossing)
    dWavFrac = pObj->dWaveLength - fmod((double)(pObj->dwCountSamples + pObj->dwCalculatedFallTimeSamples), pObj->dWaveLength);
    dwSamplesToZeroCrossing = (DWORD)floor(dWavFrac);
    if(dwSamplesToZeroCrossing == pObj->dwIntegralWaveLength)
    { 
       dwSamplesToZeroCrossing = 0;
    } // End if(dWavFrac >= 0.5)
    pObj->dwTotalFallTimeSamples = (pObj->dwCalculatedFallTimeSamples + dwSamplesToZeroCrossing);
 }else
 {
    // Total fall time samples = (Num sample to next full wave zero crossing)
    dWavFrac = pObj->dWaveLength - fmod((double)(pObj->dwCountSamples), pObj->dWaveLength);
    dwSamplesToZeroCrossing = (DWORD)floor(dWavFrac);
    if(dwSamplesToZeroCrossing == pObj->dwIntegralWaveLength)
    { 
       dwSamplesToZeroCrossing = 0;
    } // End if(dWavFrac >= 0.5)
    pObj->dwTotalFallTimeSamples = dwSamplesToZeroCrossing;
 } // End if(pObj->bUsingDecayDelay)
}

// Start the tail and continue until amplitude falls to zero.
// May or may not change the current value of caller's amplitude by side effect.
void WrapUp(ASIO_IO *pObj, double *pAmplitude)
{
 if(pObj->dwCountFallTimeSamples == 0)
 {
    CalculateTotalFallTimeSamples(pObj);
 }

 ++pObj->dwCountFallTimeSamples;
 if(pObj->dwCountFallTimeSamples >= pObj->dwTotalFallTimeSamples) // dwTotalFallTimeSamples could be 0
 {
    // Check for zero crossing
    if(sin(pObj->dPhaseAngle) >= 0)
    {
       pObj->bKeyedOn = pObj->bFalling = false;
	   // DebugFloatValue("Actual vs commanded duration in waves", ((double)pObj->dwCountSamples - (double)(pObj->dwTotalDurationSamples + pObj->dwCountFallTimeSamples)) / pObj->dWaveLength, 5);
/*
       // Debug only! Verify zero crossing...
	   if(sin(pObj->dPhaseAngle - pObj->dRadians) >= 0)
	   {
	      DebugFloatValue("Now sin(dPhaseAngle))", sin(pObj->dPhaseAngle), 5);
	      DebugFloatValue("Prev sin(dPhaseAngle)", sin(pObj->dPhaseAngle - pObj->dRadians), 5);
	      DebugFloatValue("Prev dRawSample", *pAmplitude * sin(pObj->dPhaseAngle - pObj->dRadians), 5);
	   }
*/
	   *pAmplitude = 0;
    }
 }else
 {
	if(pObj->bUsingDecayDelay)
	{
       // Raised cosine tail...
	   *pAmplitude = pObj->dAmplitude * (-0.5 * cos(TWO_PI * (double)(pObj->dwTotalFallTimeSamples - pObj->dwCountFallTimeSamples) / ((double)((pObj->dwTotalFallTimeSamples * 2) - 1))) + 0.5);
	}
 }
}

// This is called from a thread in ASIOHost that runs at THREAD_PRIORITY_HIGHEST
void SendOutgoingAudio(LPBYTE pSndBuf, DWORD dwBufsizeBytes, DWORD dwInstanceData)
{
 ASIO_IO *pObj = (ASIO_IO *)dwInstanceData;
 short *pLeftChnOut = (short *)pSndBuf;
 short *pRightChnOut = (short *)(pSndBuf + dwBufsizeBytes);
 int numSamples = dwBufsizeBytes / 2;

 if(!pObj->bKeyedOn)
 {
    ZeroMemory(pSndBuf, dwBufsizeBytes * 2);
	return;
 }

 double amplitude = pObj->dAmplitude;
 for(int n = 0; n < numSamples; n++)
 {
      pObj->dPhaseAngle += pObj->dRadians;

	  if(pObj->bRising)
	  {
		 ++pObj->dwCountRiseTimeSamples;
		 if(pObj->dwCountRiseTimeSamples >= pObj->dwTotalRiseTimeSamples) // dwTotalRiseTimeSamples could be 0
		 {
            pObj->bRising = false;
		 }else
		 {
            // Raised cosine taper
			amplitude = pObj->dAmplitude * (-0.5 * cos(TWO_PI * pObj->dwCountRiseTimeSamples / ((double)((pObj->dwTotalRiseTimeSamples * 2) - 1))) + 0.5);
	        // DebugFloatValue("amplitude", amplitude, 5); // Could fill up the debug log quickly!
		 }
	  }else if(pObj->bCheckSampleCount)
	  {
         if(pObj->dwCountSamples >= pObj->dwTotalDurationSamples)
		 {
            pObj->bCheckSampleCount = false;
            pObj->bFalling = true;
			WrapUp(pObj, &amplitude);
		 }
	  }else if(pObj->bFalling)
	  {
	 	 WrapUp(pObj, &amplitude);
	  } // End if(pObj->bFalling)

      double dRawSample = (amplitude * sin(pObj->dPhaseAngle));
	  // DebugFloatValue("dRawSample", dRawSample, 5); // Could fill up the debug log quickly!
      *pLeftChnOut = (short)_copysign(floor(fabs(dRawSample)), dRawSample);
	  *pRightChnOut++ = *pLeftChnOut++;
	  ++pObj->dwCountSamples;
 }
 ++pObj->dwCountBuffers;
}



ASIO_IO::ASIO_IO(HWND hWnd) : bASIOInitialized(false), bCreatedStereoVirtualChn(false)
{
 int numdevs = 0;
 if(!ShowASIODevices(&numdevs))return;
 if(CParams::iASIODevNum >= numdevs)
 {
    DebugMessage("Invalid ASIO device number!");
	return;
 }

 // We have ASIO devices. Go ahead and initialize ASIO for desired device...
 DebugValue("Initializing device", CParams::iASIODevNum);
 bASIOInitialized = ASIOInit(hWnd, CParams::iASIODevNum);
 if(!bASIOInitialized)
 {
    DebugMessage("Can't initialize ASIO device!");
    return;
 }
 DebugMessage("ASIO Initialized");

#ifdef USE_PROFESSIONAL_VERSION
 if(!ASIOSetSampleRate((DWORD)CParams::iSampleRate))
 {
    DebugValue("Failed, Set sample rate", CParams::iSampleRate);
 }else
 {
    DebugValue("Set sample rate", CParams::iSampleRate);
 }
#endif

 if(!ASIOCreateBuffers(NUM_STEREO_CHANNELS_REQD, STEREO_PLAY_ONLY))
 {
    DebugMessage("Failed creating ASIO buffers!");
	return;
 }


#ifdef USE_PROFESSIONAL_VERSION
 dwBufferSampleSize = ASIOGetPrimaryBufferSizeSamples();
#else
 dwBufferSampleSize = 2 * ASIOGetPrimaryBufferSizeSamples();
#endif
 DebugMessage("Create ASIO buffers success!");
 DebugValue("Buffer sample size", dwBufferSampleSize);
 dwSampleRate = ASIOGetSampleRate();
 DebugValue("Sample rate", dwSampleRate);

 InitializeToneGenerator();

 PrepareVirtualChannels();
 if(bCreatedStereoVirtualChn)
 {
    DebugMessage("Create virtual channel success!");
 }else
 {
	DebugMessage("Failed: Create virtual channel!");
 }

 dwCountBuffers = 0; // dbg only

 PreLoadBuffers(0);
 StartPlay(0);
}

ASIO_IO::~ASIO_IO()
{
}



void ASIO_IO::Shutdown(void)
{
  if(bCreatedStereoVirtualChn)
     CloseVirtualChannel(0);
  ASIOShutdown();
  DebugMessage("Shut down ASIO");
  DebugValue("Count buffers played", dwCountBuffers);
}

void ASIO_IO::PrepareVirtualChannels(void)
{
#ifdef USE_PROFESSIONAL_VERSION
 SetMaxVirtualBuffers(1);
#endif
 ZeroMemory(&StereoChannelVars, sizeof(ChannelVars));
 StereoChannelVars.dwInstanceData = (DWORD)this;
 StereoChannelVars.nVirtualChnNum = 0;
 StereoChannelVars.nBufSizeSamples = dwBufferSampleSize;
 StereoChannelVars.pOutBufServiceProc = SendOutgoingAudio;
 StereoChannelVars.pInBufServiceProc = NULL;
 bCreatedStereoVirtualChn = CreateVirtualChannel(&StereoChannelVars);
}

bool ASIO_IO::ShowASIODevices(int *pNumDevs)
{
 #define DEVNAMESSIZE 1024
 char StrBuf[DEVNAMESSIZE];
 memset(StrBuf, 0, DEVNAMESSIZE);

 int devs = ASIOGetDevices(StrBuf, DEVNAMESSIZE);
 if(!devs)
 {
    DebugMessage("No ASIO device found!\n");
    return false;
 }
 *pNumDevs = devs;

 DebugValue("ASIO devices found", devs);

 char *ptr = StrBuf;
 for(int i = 0; i < devs; i++)
 {
	 DebugValueNoCR("Device number", i);
	 DebugMessage(ptr);
     ptr += strlen(ptr);
     ++ptr;
 }
 DebugMessage("");
 DebugMessage("");

 return true;
}

void ASIO_IO::InitializeToneGenerator(void)
{
 dbLevel = MAX_16BIT_DBS - CParams::dAttenuation;
 dAmplitude = pow(10, (dbLevel / 20.0));
 dPhaseAngle = 0;
 dRadians = TWO_PI / (dwSampleRate / CParams::dFrequency);
 dWaveLength = dwSampleRate / CParams::dFrequency;
 dwIntegralWaveLength = (DWORD)floor(dWaveLength);
 dwHalfIntegralWaveLength = (DWORD)floor(dWaveLength / 2.0);
 dwCountSamples = 0;
 DebugFloatValue("WaveLength", dWaveLength, 5);

 // What is the duration of a single sample in microseconds?
 dSampleMicroseconds = 1000000.0 / dwSampleRate;


 double dNumHalfWaves;
 if(CParams::dRiseTime > 0)
 {
    dwTotalRiseTimeSamples = (DWORD)floor(CParams::dRiseTime * dwSampleRate);
    // Adjust rise time samples to nearest integral number of 1/2 dWaveLengths
    dNumHalfWaves = (double)(2 * dwTotalRiseTimeSamples) / dWaveLength;
    dwTotalRiseTimeSamples = (DWORD)floor(floor(dNumHalfWaves + 0.5) * dWaveLength / 2.0);
	DebugValue("dwTotalRiseTimeSamples", dwTotalRiseTimeSamples);
 }else dwTotalRiseTimeSamples = 0;

 if(CParams::dFallTime > 0)
 {
    dwCalculatedFallTimeSamples = (DWORD)floor(CParams::dFallTime * dwSampleRate);
    // Adjust fall time samples to nearest integral number of 1/2 dWaveLengths
    dNumHalfWaves = (double)(2 * dwCalculatedFallTimeSamples) / dWaveLength;
    dwCalculatedFallTimeSamples = (DWORD)floor(floor(dNumHalfWaves + 0.5) * dWaveLength / 2.0);
	DebugValue("dwCalculatedFallTimeSamples", dwCalculatedFallTimeSamples);
 }else dwCalculatedFallTimeSamples = 0;

 if(dwCalculatedFallTimeSamples)
 {
	bUsingDecayDelay = true;
    // nTotalFallTimeSamples will take count in nCalculatedFallTimeSamples
	// to nearest full wave zero crossing each time in KeyOff()
 }else
 {
	bUsingDecayDelay = false;
 }
 dwCountRiseTimeSamples = 0;
 dwCountFallTimeSamples = 0;
 bKeyedOn = bRising = bFalling = false;
}