// Author: George W. Taylor
// http://www.tropicalcoder.com/
// Date: 15 Oct 2012
// License: Creative Commons - Non-Commercial Use
// Link with ASIOHost.lib

#include "stdafx.h"
#include <windows.h>   
#include "CParams.h"
#include "ASIOSupport.h"
#include "PortMonitor.h"
#include "resource.h"
#include "Debug.h"

// Dimensions of the screen image...
#define SCREEN_WIDTH 330
#define SCREEN_HEIGHT 247

///////////////////////GLOBALS ////////////////////////////////////

static char *szApplicationName = "SoftCW v1.07";
static char *szWindowClassName = "SoftCW";
static CParams g_Params;
static ASIO_IO *pASIO_IO = NULL;
static PortMonitor *pPortMonitor = NULL;
static HANDLE hThread = NULL;
static HANDLE hThreadMutex = NULL;
static HINSTANCE hInstApp = NULL;
static HWND    hWndApp;
static RECT    rKeyLamp;
static HDC	   hdcScrn;
static bool    bKeyedOn = false;
static bool    bThreadLive = false;
static bool    bCommStateToggle = false;

DWORD WINAPI PortMonitorThread(LPVOID pObj);
void StartPortMonitorThread(void);
void WaitPortMonitorThreadExit(void);

extern void TimerInit(void);
extern void TimerStart(void);
extern void TimerEnd(void);
extern void TimerAverageInterval(void);
extern DWORD TimerGetAveragedInterval(void);
extern DWORD TimerGetInterval(void);
extern DWORD TimerGetRunCount(void);
extern void TimerEnableRecurringEvent(bool bState);
extern void TimerRecurringEvent(void);
extern void TimerShowIntervalTime(void);


//---------------------------- Cleanup ----------------------------------
//
//	Simply cleans up any memory issues when the application exits
//-----------------------------------------------------------------------
void Cleanup()
{
 if(pPortMonitor != NULL)
 {
    // Release the port monitor thread
    bThreadLive = false;
    pPortMonitor->CancelComWait();
	// Ensure port monitor thread has exited
	// before deleting PortMonitor object
    WaitPortMonitorThreadExit();
    delete pPortMonitor;
	pPortMonitor = NULL;
 }

 if(pASIO_IO != NULL)
 {
    pASIO_IO->Shutdown();
    delete pASIO_IO;
	pASIO_IO = NULL;
 }

 CloseDebugFile();
}

// Sounds the tone generator and lights the key lamp
void KeyOn(void)
{
 if(!bKeyedOn)
 {
    TimerStart();
    pASIO_IO->KeyOn();
	bCommStateToggle = bKeyedOn = true;
	Ellipse(hdcScrn, rKeyLamp.left, rKeyLamp.top, rKeyLamp.right, rKeyLamp.bottom);
 }
}

// Stops the tone generator and kills the key lamp
void KeyOff(void)
{
 if(bKeyedOn)
 {
    TimerEnd();
    pASIO_IO->KeyOff(TimerGetInterval());
    bCommStateToggle = bKeyedOn = false;
	InvalidateRect(hWndApp, &rKeyLamp, false);
	UpdateWindow(hWndApp);
 }
}

void StartPortMonitorThread(void)
{
 DWORD dwThreadID;

 hThreadMutex = CreateMutex(NULL,FALSE,NULL);
 if(hThreadMutex == NULL)
 {
    DebugMessage("Failed creating thread mutex!"); 
	return;
 }
 bThreadLive = true;
 hThread = CreateThread(NULL, 0, PortMonitorThread, (LPVOID)NULL, 0, &dwThreadID);
 if(hThread == NULL)
 {
    DebugMessage("Failed creating port monitor thread");
	bThreadLive = false;
    return;
 }
 SetThreadPriority(hThread, THREAD_PRIORITY_TIME_CRITICAL);
 DebugMessage("Started port monitor thread");
}

void WaitPortMonitorThreadExit(void)
{
 DWORD dwWaitResult;
 int nCountWaits;

 if(hThread == NULL)return;
 if(hThreadMutex == NULL)return;

 Sleep(1);
 nCountWaits = 1;
 dwWaitResult = WaitForSingleObject(hThreadMutex, 1);
 while(dwWaitResult == WAIT_TIMEOUT)
 {
       dwWaitResult = WaitForSingleObject(hThreadMutex, 1);
	   if(++nCountWaits == 250)break; // We are nor going to wait all day for the thread to exit!
 }

 DebugValue("Time waiting for thread to release mutex", nCountWaits);
 ReleaseMutex(hThreadMutex);
 CloseHandle(hThreadMutex);
 CloseHandle(hThread);
 hThread = hThreadMutex = NULL;
}

DWORD WINAPI PortMonitorThread(LPVOID pObj)
{
 int res = 0;

 if(hThreadMutex != NULL)WaitForSingleObject(hThreadMutex, INFINITE);

 while(bThreadLive && (res >= 0))
 {
	   if(pPortMonitor == NULL)break;
	   res = pPortMonitor->ReadPort();
	   if(res < 0)
	   {
		  DebugMessage("Failed: ReadPort!");
		  break;
	   }

	   if(res)
	   {
		  bCommStateToggle = pPortMonitor->LineSignaled();
	      if(bCommStateToggle)
	      {
             KeyOn();
	      }else
	      {
             KeyOff();
	      }

		  // Software switch contact debouncing...
		  int delay = CParams::iDebounceDelay;
	      while(delay)
		  {
			    if(!bThreadLive)break;
			    Sleep(1);
				--delay;
		  }
	   }
 }
 DebugMessage("Exiting port monitor thread!");
 if(hThreadMutex != NULL)ReleaseMutex(hThreadMutex);
 return 0;
}

void PreparePortMonitor(void)
{
 char PortName[32];
 char EventName[32];
 DWORD dwComEvent = 0;

 // Was a port specified?
 strcpy(PortName, g_Params.sPort);
 PortName[3] = '\0';
 if(!strcmp(PortName, "LPT") || !strcmp(PortName, "COM"))
 {
    if(!strcmp(g_Params.sComEventMonitored, "CTS"))
	{
       // We will be monitoring clear-to-send signal
	   dwComEvent = EV_CTS;
	   strcpy(EventName, "CTS");
	}else if(!strcmp(g_Params.sComEventMonitored, "DSR"))
	{
       // We will be monitoring data-set-ready signal
	   dwComEvent = EV_DSR;
	   strcpy(EventName, "DSR");
	}else if(!strcmp(g_Params.sComEventMonitored, "RIN")) // user can put "RING" in the config file but we only retain the first 3 letters
	{
       // We will be monitoring ring indicator
	   dwComEvent = EV_RING;
	   strcpy(EventName, "RING");
	}else if(!strcmp(g_Params.sComEventMonitored, "RLS")) // user can put "RLSD" in the config file but we only retain the first 3 letters
	{
       // We will be monitoring receive-line-signal-detect
	   dwComEvent = EV_RLSD;
	   strcpy(EventName, "RLSD");
	} 

    if(dwComEvent)
    {
	   pPortMonitor = new PortMonitor();
	   if(!pPortMonitor->InitPort(hWndApp, g_Params.sPort, dwComEvent))
	   {
          delete pPortMonitor;
		  pPortMonitor = NULL;
		  wsprintf(PortName, "Failed opening port: %s", g_Params.sPort);
          DebugMessage(PortName);
		  return;
	   } 
	}
 }

 if(!dwComEvent)
 {
    DebugMessage("Invalid port params!");
 }else
 {
    // It's a go! 
    wsprintf(PortName, "Monitoring port %s for %s", g_Params.sPort, EventName);
    DebugMessage(PortName);
	DebugValue("Debounce delay", CParams::iDebounceDelay);
    StartPortMonitorThread();
 }
}


//-----------------------------------WinProc-----------------------------
//
//-----------------------------------------------------------------------
LRESULT CALLBACK WindowProc(HWND hWnd, 
						    UINT msg, 
                            WPARAM wparam, 
                            LPARAM lparam)
{
 static HDC		hdcBackBuffer;
 static HBITMAP	hBitmap;
 static HBITMAP	hOldBitmap; 
 static HBRUSH  hOldBrush = NULL;
 static HANDLE  hCurThread;
 static DWORD   dwThreadPriority;
 static int cxClient, cyClient;


 switch(msg)
 {	
		case WM_CREATE: 
		{

			// Get the size of the client window
			RECT rect;
			GetClientRect(hWnd, &rect);

			cxClient = rect.right;
			cyClient = rect.bottom;

			// Location on screen of the key lamp...
			rKeyLamp.left = 10;
			rKeyLamp.top = cyClient - 30;
			rKeyLamp.right = rKeyLamp.left + 20;
			rKeyLamp.bottom = rKeyLamp.top + 20;

			// Create a surface for the backbuffer
			hdcBackBuffer = CreateCompatibleDC(NULL);
			hdcScrn = GetDC(hWnd);
            hOldBrush = (HBRUSH)SelectObject(hdcScrn, CreateSolidBrush(0x0000FF00)); 

			// Load the screen image
            hBitmap = (HBITMAP)LoadImage(hInstApp, MAKEINTRESOURCE(IDB_BITMAP1), IMAGE_BITMAP, 0, 0, LR_DEFAULTCOLOR);

			// Select it into the back buffer
			hOldBitmap = (HBITMAP)SelectObject(hdcBackBuffer, hBitmap); 
			InitDebugFile();
			DebugMessage(szApplicationName);
			DebugMessage("");
			pASIO_IO = new ASIO_IO(hWnd);
			DebugFloatValue("Frequency", CParams::dFrequency, 5);
			DebugFloatValue("Attenuation", CParams::dAttenuation, 5);
	        DebugFloatValue("Rise time", CParams::dRiseTime, 5);
	        DebugFloatValue("Fall time", CParams::dFallTime, 5);
            PreparePortMonitor();
			TimerInit();
		}  
			
		break;
		
		// Check key press messages
		case WM_KEYUP:
		{
			switch(wparam)
			{
				case VK_ESCAPE:
			  	 PostQuitMessage(0);
				break;

				case VK_SPACE:
                 hCurThread = GetCurrentThread();
                 dwThreadPriority = GetThreadPriority(hCurThread);
                 if(SetThreadPriority(hCurThread, THREAD_PRIORITY_TIME_CRITICAL))
				 {
				    KeyOff();
                    SetThreadPriority(hCurThread, dwThreadPriority);
				 }
				break;
			} // end WM_KEYUP switch
		}
		break;
		case WM_KEYDOWN:
		{
			switch(wparam)
			{
				case VK_SPACE:
                 hCurThread = GetCurrentThread();
                 dwThreadPriority = GetThreadPriority(hCurThread);
                 if(SetThreadPriority(hCurThread, THREAD_PRIORITY_TIME_CRITICAL))
				 {  
				    KeyOn();
                    SetThreadPriority(hCurThread, dwThreadPriority);
				 }
				break;
			} // end WM_KEYDOWN switch
		}
		break;

		case WM_PAINT: 
		{
         PAINTSTRUCT ps;
      
		  BeginPaint(hWnd, &ps);
		  if(!bKeyedOn)
		  {
		      // Blit backbuffer to front
		      BitBlt(ps.hdc, ps.rcPaint.left, ps.rcPaint.top, ps.rcPaint.right, ps.rcPaint.bottom, hdcBackBuffer, ps.rcPaint.left, ps.rcPaint.top, SRCCOPY); 
		  }
		  EndPaint(hWnd, &ps);
		} 
			
		break;
		case WM_WANTS_RESET:
         // A device may request a reset at any time.
         // Some sound cards require a complete reset when an application requests a change in sample rate
         // In this case, ASIOHost will send a WM_WANTS_RESET message, as defined in constants.h
         MessageBox(NULL, "Please exit the application and restart it for change to take effect.",
                    "Sample rate change request...", MB_OK | MB_ICONEXCLAMATION);
	    break;

		case WM_DESTROY: 
		{
			SelectObject(hdcBackBuffer, hOldBitmap);
            SelectObject(hdcScrn, hOldBrush); 
			
			// Clean up our backbuffer objects
			DeleteDC(hdcBackBuffer);
			DeleteObject(hBitmap); 

            // Kill the application. This sends a WM_QUIT message 
			PostQuitMessage(0);
 		} 
			
		break;

		default:break;

	} // end switch

	// default msg handler 
	return (DefWindowProc(hWnd, msg, wparam, lparam));

}//end WinProc


//-----------------------------------WinMain-----------------------------------------
//	Entry point for our windows application
//-----------------------------------------------------------------------------------
int WINAPI WinMain(HINSTANCE hinstance,
				   HINSTANCE hprevinstance,
				   LPSTR lpcmdline,
				   int ncmdshow)
{

	WNDCLASSEX winclass; 
	MSG		   msg;		 

	// first fill in the window class stucture
	winclass.cbSize        = sizeof(WNDCLASSEX);
	winclass.style		   = CS_HREDRAW | CS_VREDRAW;
	winclass.lpfnWndProc   = WindowProc;
	winclass.cbClsExtra	   = 0;
	winclass.cbWndExtra	   = 0;
	winclass.hInstance	   = hinstance;
	winclass.hIcon		   = LoadIcon(hinstance, MAKEINTRESOURCE(IDI_ICON1));
	winclass.hCursor	   = LoadCursor(NULL, IDC_ARROW); 
	winclass.hbrBackground = NULL; 
	winclass.lpszMenuName  = NULL;
	winclass.lpszClassName = szWindowClassName;
	winclass.hIconSm       = LoadIcon(hinstance, MAKEINTRESOURCE(IDI_ICON1));

    hInstApp = hinstance;

	// Register the window class
	if(!RegisterClassEx(&winclass))
	{
		MessageBox(NULL, "Error Registering Class!", "Error", 0);
        return 0;
	}

  RECT rApp, rDeskTop;
  ZeroMemory(&rDeskTop, sizeof(RECT));
  GetClientRect(GetDesktopWindow(), &rDeskTop);

  int xborder = 2 * GetSystemMetrics(SM_CXFRAME); // 2 * 8
  int yborder = 2 * GetSystemMetrics(SM_CYFRAME); // 2 * 8
  int caption = GetSystemMetrics(SM_CYCAPTION); // 22
  int menu = 0; // GetSystemMetrics(SM_CYMENU); // 20
  int width = SCREEN_WIDTH + xborder;
  int height = SCREEN_HEIGHT + yborder + caption + menu;
  int x = (rDeskTop.right / 2) - (width / 2);
  int y = (rDeskTop.bottom / 2) - (height / 2);

  hWndApp = CreateWindow(szWindowClassName, szApplicationName, WS_OVERLAPPEDWINDOW | WS_VISIBLE,
                         x, y, width, height, NULL, NULL, hinstance, NULL);
  ZeroMemory(&rApp, sizeof(RECT));
  GetClientRect(hWndApp, &rApp);
	
  // Show the window
  ShowWindow(hWndApp, SW_SHOWDEFAULT );
  UpdateWindow(hWndApp);

  // Main message loop:
 while(GetMessage(&msg, NULL, 0, 0))
 {
	   TranslateMessage(&msg);
	   DispatchMessage(&msg);
 }

 // Clean up everything and exit the app
 Cleanup();
 return (int) msg.wParam;
}

